/*
Parrot
Copyright (C) 2003 Douglas Thain
http://www.cs.wisc.edu/~thain/research/parrot
This program is released under the GNU General Public License.
See the file COPYING for details.
*/

#include "pfs_resolve.h"
#include "pfs_types.h"

#include "chirp_client.h"
#include "debug.h"
#include "stringtools.h"
#include "xmalloc.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/wait.h>

#define PFS_NAME_FLAGS_SLOPPYCACHE 1
#define PFS_NAME_FLAGS_STRICTCACHE 2

extern char *pfs_process_name();
extern void pfs_process_kill();
extern int pfs_master_timeout;

struct mount_entry {
	char prefix[PFS_PATH_MAX];
	char redirect[PFS_PATH_MAX];
	int flags;
	struct mount_entry *next;
};

static int chirp_enable = 0;
static struct chirp_client * chirp_resolver = 0;
static struct mount_entry * mountlist = 0;

void pfs_resolve_mountlist_config( const char *prefix, const char *redirect, int fl )
{
	struct mount_entry *e = xmalloc(sizeof(*e));
	strcpy(e->prefix,prefix);
	strcpy(e->redirect,redirect);
	e->flags = fl;
	e->next = mountlist;
	mountlist = e;
}

static char * lookup_env( const char *name, void *arg )
{
	char *value = getenv(name);
	if(value) {
		return strdup(value);
	} else {
		fatal("mountfile contains an undefined variable: %s\n",name);
		return 0;
	}
}

void pfs_resolve_mountfile_config( const char *mountfile )
{
	char line[PFS_LINE_MAX];
	char prefix[PFS_LINE_MAX];
	char redirect[PFS_LINE_MAX];
	char flagname[PFS_LINE_MAX];
	int fields;
	FILE *file;
	int linenum=0;
	char *cooked;

	file = fopen(mountfile,"r");
	if(!file) fatal("couldn't open mountfile %s: %s\n",mountfile,strerror(errno));

	while(1) {
		if(!fgets(line,sizeof(line),file)) {
			if(errno==EINTR) {
				continue;
			} else {
				break;
			}
		}
		linenum++;
		if(line[0]=='#') continue;
		string_chomp(line);
		if(!line[0]) continue;

		cooked = xstrdup(line);
		if(!cooked) fatal("out of memory");
		cooked = string_subst(cooked,lookup_env,0);
		fields = sscanf(cooked,"%s %s %s",prefix,redirect,flagname);
		free(cooked);

		if(fields==0) {
			continue;
		} else if(fields<2) {
			fatal("%s has an error on line %d\n",mountfile,linenum);
		} else {
			int flags = 0;
			if(fields>2) {
				if(!strcmp(flagname,"cache")) {
					flags = PFS_NAME_FLAGS_STRICTCACHE;
				} else if(!strcmp(flagname,"sloppycache")) {
					flags = PFS_NAME_FLAGS_SLOPPYCACHE;
				} else {
					fatal("%s: has an unknown keyword '%s' on line %d\n",mountfile,flagname,linenum);
				}
			}
			pfs_resolve_mountlist_config(prefix,redirect,flags);
		}
	}

	fclose(file);
}

void pfs_resolve_mountstring_config( const char *string )
{
	char *cstring = xstrdup(string);
	char *t;
	char logical[PFS_PATH_MAX];
	char physical[PFS_PATH_MAX];

	t = strtok(cstring,",");
	while(t) {
		if(sscanf(t,"%[^=]=%s",logical,physical)==2) {
			pfs_resolve_mountlist_config(logical,physical,0);
		} else {
			fatal("syntax error in mountstring: %s",t);
		}
		t = strtok(0,",");
	}

	free(cstring);
}


void pfs_resolve_chirp_config()
{
	chirp_enable = 1;
}

/*
Run an external program in order to pop up a window
that asks the user if they want to permit access to this file.
*/

static pfs_resolve_t ask_user( const char *logical_name )
{
	pfs_resolve_t result;
	int save_errno;
	int exitstatus;
	int exitcode;
	FILE *stream;
	stream = popen("xmessage -buttons Yes:10,No:11,Kill:12 -default No -nearmouse -file -","w");
	if(stream) {
		fprintf(stream,"Notice from PFS:\n");
		fprintf(stream,"   %s\n",pfs_process_name());
		fprintf(stream,"Wants access to the file:\n");
		fprintf(stream,"   %s\n",logical_name);
		fprintf(stream,"Should it be permitted to access it?\n");
		exitstatus = pclose(stream);
		save_errno = errno;
		if(WIFEXITED(exitstatus)) {
			exitcode = WEXITSTATUS(exitstatus);
		} else {
			exitcode = 0;
		}
	} else {
		exitcode = 1;
	}

	if(exitcode==10) {
		result = PFS_RESOLVE_UNCHANGED;
	} else if(exitcode==11) {
		result = PFS_RESOLVE_DENIED;
	} else if(exitcode==12) {
		result = PFS_RESOLVE_DENIED;
		pfs_process_kill();
	} else {
		debug(D_RESOLVE|D_NOTICE,"wanted to ask user about access to %s",logical_name);
		debug(D_RESOLVE|D_NOTICE,"but couldn't invoke xmessage: %s",strerror(save_errno));
		result = PFS_RESOLVE_DENIED;
		pfs_process_kill();
	}

	return result;
}

/*
Compare a logical name to a mountlist entry and
determine what to do with it.
*/

static pfs_resolve_t mount_entry_check( const char *logical_name, const char *prefix, const char *redirect, char *physical_name )
{
	pfs_resolve_t result;

	int plen = strlen(prefix);
	int llen = strlen(logical_name);
	if(
		!strncmp(prefix,logical_name,plen) &&
		(
			prefix[plen-1]=='/' ||
			logical_name[plen]=='/' ||
			plen==llen
		)
	) {
		if(!strcmp(redirect,"DENY")) {
			result = PFS_RESOLVE_DENIED;
		} else if(!strcmp(redirect,"ASK")) {
			result = ask_user(logical_name);
		} else {
			strcpy(physical_name,redirect);
			if(llen>plen) strcat(physical_name,&logical_name[plen]);
			result = PFS_RESOLVE_CHANGED;
		}
	} else {
		result = PFS_RESOLVE_UNCHANGED;
	}

	return result;
}

/* Search for mount entries in the internal mountlist. */

static pfs_resolve_t resolve_mountlist( const char *logical_name, char *physical_name, int *flags )
{
	struct mount_entry *e;
	pfs_resolve_t result = PFS_RESOLVE_UNCHANGED;

	*flags = 0;

	for(e=mountlist;e;e=e->next) {
		result = mount_entry_check(logical_name,e->prefix,e->redirect,physical_name);
		if(result!=PFS_RESOLVE_UNCHANGED) {
			*flags = e->flags;
			break;
		}
	}

	return result;
}

/* Ask Chirp to resolve a name. */

static pfs_resolve_t resolve_chirp( const char *logical_name, char *physical_name )
{
	char *url;
	char prefix[PFS_PATH_MAX];
	char tmp[PFS_PATH_MAX];
	int length;

	if(!chirp_enable) return PFS_RESOLVE_UNCHANGED;

	if(!chirp_resolver) {
		chirp_resolver = chirp_client_connect_condor(time(0)+pfs_master_timeout);
		if(!chirp_resolver) {
			fatal("couldn't connect to Condor Chirp proxy: %s",strerror(errno));
		}
	}

	length = chirp_client_lookup(chirp_resolver,logical_name,&url,time(0)+pfs_master_timeout);
	if(length>0) {
		strncpy(physical_name,url,length);
		physical_name[length] = 0;
		free(url);

		/*
		The Condor chirp names come in a somewhat different form.
		They look like buffer:remote:/foo/bar and local:/biz/boo.
		*/

		while(1) {
			if(!strncmp(physical_name,"buffer:",7)) {
				strcpy(tmp,physical_name+7);
				strcpy(physical_name,tmp);
			} else {
				if(sscanf(physical_name,"%[^:]:",prefix)) {
					int plen = strlen(prefix);
					if(!strcmp(prefix,"remote")) {
						strcpy(prefix,"chirp/CONDOR");
					}
					sprintf(tmp,"/%s/%s",prefix,physical_name+plen+2);
					strcpy(physical_name,tmp);
				} else {
					/* leave it as is */
				}
				return PFS_RESOLVE_CHANGED;
			}
		}
	} else {
		fatal("couldn't resolve name via Condor Chirp proxy: %s",strerror(errno));
	}

	return PFS_RESOLVE_DENIED;
}

pfs_resolve_t pfs_resolve( const char *logical_name, char *physical_name, int *flags )
{
	pfs_resolve_t result = PFS_RESOLVE_UNCHANGED;

	result = resolve_mountlist(logical_name,physical_name,flags);
	if(result==PFS_RESOLVE_UNCHANGED)
		result = resolve_chirp(logical_name,physical_name);

	switch(result) {
		case PFS_RESOLVE_UNCHANGED:
			debug(D_RESOLVE,"%s unchanged",logical_name);
			strcpy(physical_name,logical_name);
			break;
		case PFS_RESOLVE_CHANGED:
			debug(D_RESOLVE,"%s becomes %s",logical_name,physical_name);
			break;
		case PFS_RESOLVE_DENIED:
			debug(D_RESOLVE|D_NOTICE,"%s DENIED",logical_name);
			break;
	}

	return result;
}


