/*
PFS is Copyright (C) 2001 Douglas Thain.
This program is released under the GNU General Public License.
See the file COPYING for details.
*/

#ifdef HAS_DCAP

#include "pfs_service.h"
#include "pfs_file.h"

/*
Currently, there is a conflict between debug()
defined by libdttools.a and debug() defined by libdcap.
For the moment, we hack around it.
*/

extern "C" {
#include "debug.h"
#define debug dcap_debug
#include "dcap.h"
#undef debug
}

#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

class pfs_file_dcap : public pfs_file
{
private:
	int fd;
	pfs_off_t remote_offset;

public:
	pfs_file_dcap( pfs_name *n, int f ) : pfs_file(n) {
		fd = f;
		remote_offset = 0;
	}
	virtual int close() {
		int result;
		debug(D_DCAP,"close %d",fd);
		int fl = debug_flags_clear();
		result = ::dc_close(fd);
		debug_flags_restore(fl);
		debug(D_DCAP,"= %d %s",result,result>=0?"":strerror(errno));
		return result;
	}
	virtual pfs_ssize_t read( void *data, pfs_size_t length, pfs_off_t offset ) {
		int result;
		debug(D_DCAP,"pread %d %d %d",fd,length,offset);
		int fl = debug_flags_clear();
		result =::dc_pread(fd,(char*)data,length,offset);
		debug_flags_restore(fl);
		debug(D_DCAP,"= %d %s",result,result>=0?"":strerror(errno));
		return result;
	}
	virtual pfs_ssize_t write( const void *data, pfs_size_t length, pfs_off_t offset ) {
		int result;
		debug(D_DCAP,"pwrite %d %d %d",fd,length,offset);
		int fl = debug_flags_clear();
		result =::dc_pwrite(fd,(char*)data,length,offset);
		debug_flags_restore(fl);
		debug(D_DCAP,"= %d %s",result,result>=0?"":strerror(errno));
		return result;
	}
	virtual pfs_ssize_t get_size() {
		pfs_ssize_t result;
		result = ::dc_lseek(fd,0,SEEK_END);
		if(result>=0) {
			remote_offset = result;
			return result;
		} else {
			return 0;
		}
	}
};

class pfs_service_dcap : public pfs_service {
public:
	virtual int get_default_port() {
		return 22125;
	}

	virtual pfs_file * open( pfs_name *name, int flags, mode_t mode ) {
		int result;
		char url[PFS_PATH_MAX];
		sprintf(url,"dcap://%s:%d/%s",name->host,name->port,name->rest);
		debug(D_DCAP,"open %s %d %d",url,flags,mode);
		int fl = debug_flags_clear();
		result = dc_open(url,flags,mode);
		debug_flags_restore(fl);
		debug(D_DCAP,"= %d %s",result,result>=0?"":strerror(errno));

		if(result>=0) {
			dc_unsafeWrite(result);
			dc_noBuffering(result);
			return new pfs_file_dcap(name,result);
		} else {
			return 0;
		}
	}

	virtual int stat( pfs_name *name, struct stat *buf )
	{
		int result;
		char url[PFS_PATH_MAX];
		sprintf(url,"dcap://%s:%d/%s",name->host,name->port,name->rest);
		debug(D_DCAP,"stat %s",url);
		int fl = debug_flags_clear();
		result = dc_stat(url,buf);
		debug_flags_restore(fl);
		debug(D_DCAP,"= %d %s",result,result>=0?"":strerror(errno));
		return result;
	}

	virtual int lstat( pfs_name *name, struct stat *buf )
	{
		int result;
		char url[PFS_PATH_MAX];
		sprintf(url,"dcap://%s:%d/%s",name->host,name->port,name->rest);
		debug(D_DCAP,"lstat %s",url);
		int fl = debug_flags_clear();
		result = dc_lstat(url,buf);
		debug_flags_restore(fl);
		debug(D_DCAP,"= %d %s",result,result>=0?"":strerror(errno));
		return result;
	}


	virtual int is_seekable() {
		return 1;
	}
};

static pfs_service_dcap pfs_service_dcap_instance;
pfs_service *pfs_service_dcap = &pfs_service_dcap_instance;

#endif
